"use strict";
var __extends = (this && this.__extends) || function (d, b) {
    for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p];
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
};
var objectAssign = require('object-assign');
function isTypeDeclaration(fileName) {
    return /\.d.ts$/.test(fileName);
}
exports.isTypeDeclaration = isTypeDeclaration;
function isImportOrExportDeclaration(node) {
    return (!!node.exportClause || !!node.importClause)
        && node.moduleSpecifier;
}
function isImportEqualsDeclaration(node) {
    return !!node.moduleReference && node.moduleReference.hasOwnProperty('expression');
}
var FileAnalyzer = (function () {
    function FileAnalyzer(state) {
        this.dependencies = new DependencyManager();
        this.validFiles = new ValidFilesManager();
        this.state = state;
    }
    FileAnalyzer.prototype.checkDependencies = function (fileName, isDefaultLib) {
        if (isDefaultLib === void 0) { isDefaultLib = false; }
        var isValid = this.validFiles.isFileValid(fileName);
        if (isValid) {
            return isValid;
        }
        this.validFiles.markFileValid(fileName);
        this.dependencies.clearDependencies(fileName);
        var changed = false;
        try {
            if (!this.state.hasFile(fileName)) {
                this.state.readFileAndAdd(fileName, isDefaultLib);
                changed = true;
            }
            this.checkDependenciesInternal(fileName);
        }
        catch (err) {
            this.validFiles.markFileInvalid(fileName);
            throw err;
        }
        return changed;
    };
    FileAnalyzer.prototype.checkDependenciesInternal = function (fileName) {
        var _this = this;
        var imports = this.findImportDeclarations(fileName);
        imports.forEach(function (imp) {
            _this.dependencies.addDependency(fileName, imp);
            _this.checkDependencies(imp.resolvedFileName);
        });
        return null;
    };
    FileAnalyzer.prototype.findImportDeclarations = function (fileName) {
        var _this = this;
        var sourceFile = this.state.getSourceFile(fileName);
        var isDeclaration = isTypeDeclaration(fileName);
        var imports = [];
        var visit = function (node) {
            if (!isDeclaration && isImportEqualsDeclaration(node)) {
                var importPath = node.moduleReference.expression.text;
                imports.push(importPath);
            }
            else if (!isDeclaration && isImportOrExportDeclaration(node)) {
                var importPath = node.moduleSpecifier.text;
                imports.push(importPath);
            }
        };
        imports.push.apply(imports, sourceFile.referencedFiles.map(function (file) { return file.fileName; }));
        this.state.ts.forEachChild(sourceFile, visit);
        var resolvedImports = imports.map(function (importPath) {
            return _this.resolve(fileName, importPath);
        });
        return resolvedImports.filter(Boolean);
    };
    FileAnalyzer.prototype.resolve = function (fileName, depName) {
        if (/^[a-z0-9].*\.d\.ts$/.test(depName)) {
            depName = './' + depName;
        }
        var resolution = this.state.ts.resolveModuleName(depName, fileName, this.state.compilerConfig.options, this.state.ts.sys);
        var resolvedModule = resolution.resolvedModule;
        if (resolvedModule) {
            this.state.fileAnalyzer.dependencies.addResolution(fileName, depName, resolvedModule);
        }
        if (this.state.compilerConfig.options.traceResolution) {
            console.log('Resolve', depName, '\n', 'from', fileName, '\n', 'resolved', resolvedModule && resolvedModule.resolvedFileName, '\n', 'failedLookupLocations', resolution.failedLookupLocations, '\n\n');
        }
        return resolvedModule;
    };
    return FileAnalyzer;
}());
exports.FileAnalyzer = FileAnalyzer;
var DependencyManager = (function () {
    function DependencyManager() {
        this.dependencies = {};
        this.knownTypeDeclarations = {};
        this.compiledModules = {};
        this.resolutions = {};
    }
    DependencyManager.prototype.addResolution = function (fileName, depName, resolvedModule) {
        this.resolutions[(fileName + "::" + depName)] = resolvedModule;
    };
    DependencyManager.prototype.getResolution = function (fileName, depName) {
        return this.resolutions[(fileName + "::" + depName)];
    };
    DependencyManager.prototype.addDependency = function (fileName, dep) {
        if (!this.dependencies.hasOwnProperty(fileName)) {
            this.clearDependencies(fileName);
        }
        this.dependencies[fileName].push(dep);
    };
    DependencyManager.prototype.addCompiledModule = function (fileName, depFileName) {
        if (!this.compiledModules.hasOwnProperty(fileName)) {
            this.clearCompiledModules(fileName);
        }
        var store = this.compiledModules[fileName];
        if (store.indexOf(depFileName) === -1) {
            store.push(depFileName);
        }
    };
    DependencyManager.prototype.clearDependencies = function (fileName) {
        this.dependencies[fileName] = [];
    };
    DependencyManager.prototype.clearCompiledModules = function (fileName) {
        this.compiledModules[fileName] = [];
    };
    DependencyManager.prototype.getDependencies = function (fileName) {
        if (!this.dependencies.hasOwnProperty(fileName)) {
            this.clearDependencies(fileName);
        }
        return this.dependencies[fileName].slice();
    };
    DependencyManager.prototype.addTypeDeclaration = function (fileName) {
        this.knownTypeDeclarations[fileName] = true;
    };
    DependencyManager.prototype.hasTypeDeclaration = function (fileName) {
        return this.knownTypeDeclarations.hasOwnProperty(fileName);
    };
    DependencyManager.prototype.getTypeDeclarations = function () {
        return objectAssign({}, this.knownTypeDeclarations);
    };
    DependencyManager.prototype.getDependencyGraph = function (fileName) {
        var _this = this;
        var appliedDeps = {};
        var result = {
            fileName: fileName,
            dependencies: []
        };
        var walk = function (fileName, context) {
            _this.getDependencies(fileName).forEach(function (dep) {
                var fileName = dep.resolvedFileName;
                var depContext = {
                    fileName: dep.resolvedFileName,
                    dependencies: []
                };
                context.dependencies.push(depContext);
                if (!appliedDeps[fileName]) {
                    appliedDeps[fileName] = true;
                    walk(fileName, depContext);
                }
            });
        };
        walk(fileName, result);
        return result;
    };
    DependencyManager.prototype.applyCompiledFiles = function (fileName, deps) {
        if (!this.compiledModules.hasOwnProperty(fileName)) {
            this.clearCompiledModules(fileName);
        }
        this.compiledModules[fileName].forEach(function (mod) {
            deps.add(mod);
        });
    };
    DependencyManager.prototype.applyChain = function (fileName, deps) {
        if (!this.dependencies.hasOwnProperty(fileName)) {
            this.clearDependencies(fileName);
        }
        var appliedDeps = {};
        var graph = this.getDependencyGraph(fileName);
        var walk = function (item) {
            var itemFileName = item.fileName;
            if (!appliedDeps[itemFileName]) {
                appliedDeps[itemFileName] = true;
                deps.add(itemFileName);
                item.dependencies.forEach(function (dep) { return walk(dep); });
            }
        };
        walk(graph);
    };
    return DependencyManager;
}());
exports.DependencyManager = DependencyManager;
var ValidFilesManager = (function () {
    function ValidFilesManager() {
        this.files = {};
    }
    ValidFilesManager.prototype.isFileValid = function (fileName) {
        return this.files[fileName];
    };
    ValidFilesManager.prototype.markFileValid = function (fileName) {
        this.files[fileName] = true;
    };
    ValidFilesManager.prototype.markFileInvalid = function (fileName) {
        this.files[fileName] = false;
    };
    return ValidFilesManager;
}());
exports.ValidFilesManager = ValidFilesManager;
var ResolutionError = (function (_super) {
    __extends(ResolutionError, _super);
    function ResolutionError() {
        _super.apply(this, arguments);
    }
    return ResolutionError;
}(Error));
exports.ResolutionError = ResolutionError;
//# sourceMappingURL=deps.js.map